package core

import (
	"git.akyoto.dev/cli/q/src/asm"
	"git.akyoto.dev/cli/q/src/cpu"
	"git.akyoto.dev/cli/q/src/errors"
	"git.akyoto.dev/cli/q/src/expression"
	"git.akyoto.dev/cli/q/src/token"
	"git.akyoto.dev/cli/q/src/types"
)

// ExpressionToRegister puts the result of an expression into the specified register.
func (f *Function) ExpressionToRegister(node *expression.Expression, register cpu.Register) (types.Type, error) {
	if node.IsFolded {
		f.RegisterNumber(asm.MOVE, register, node.Value)
		return types.Int, nil
	}

	if node.IsLeaf() {
		return f.TokenToRegister(node.Token, register)
	}

	switch node.Token.Kind {
	case token.Call:
		return f.CallToRegister(node, register)
	case token.Array:
		return f.ArrayElementToRegister(node, register)
	case token.Period:
		return f.PeriodToRegister(node, register)
	}

	if len(node.Children) == 1 {
		if !node.Token.IsUnaryOperator() {
			return nil, errors.New(errors.MissingOperand, f.File, node.Token.End())
		}

		typ, err := f.ExpressionToRegister(node.Children[0], register)

		if err != nil {
			return typ, err
		}

		return typ, f.ExecuteRegister(node.Token, register)
	}

	left := node.Children[0]
	right := node.Children[1]
	final := register

	if f.UsesRegister(right, register) {
		register = f.NewRegister()
	}

	typ, err := f.ExpressionToRegister(left, register)

	if err != nil {
		return nil, err
	}

	err = f.Execute(node.Token, register, right)

	if register != final {
		f.RegisterRegister(asm.MOVE, final, register)
		f.FreeRegister(register)
	}

	_, isArray := typ.(*types.Array)

	if isArray {
		typ = types.AnyPointer
	} else if right.Token.Kind == token.Identifier {
		rightVariable := f.VariableByName(right.Token.Text(f.File.Bytes))

		if typ == types.AnyPointer && rightVariable.Type == types.AnyPointer {
			typ = types.Int
		}
	}

	return typ, err
}
